#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include "renderer.h"
#include "device-output.h"

#undef OSDEBUG

#define UCLIMIT(x) ((t = (x)), (t &= ~(t>>31)), (t | ~((t-256) >> 31)))

#define RIDX 0
#define GIDX 1
#define BIDX 2    
#define TEX_WIDTH 512
#define TEX_HEIGHT 512
#define TEX_OFFSET(x, y) ((y) * TEX_WIDTH * 3 + (x) * 3)
#define TEX_OFFSET_R(x, y) ((y) * TEX_WIDTH * 3 + (x) * 3 + RIDX)
#define TEX_OFFSET_G(x, y) ((y) * TEX_WIDTH * 3 + (x) * 3 + GIDX)
#define TEX_OFFSET_B(x, y) ((y) * TEX_WIDTH * 3 + (x) * 3 + BIDX)

#include "../include/CAVImagePool.h"
#include "../include/FLSharedMemory.h"

class ShmemOutputDevice : public OutputDevice
{
public:
  ShmemOutputDevice(const char *name);
  virtual int command(int argc, const char*const* argv);
  CAVImagePool *get_pool(void) { return (CAVImagePool *)cshm_->GetAddress(); }
  void attach_shmem(int port);
private:
  FLSharedMemory *cshm_;
};

class ShmemOutboard : public Renderer
{
public:
    ShmemOutboard(ShmemOutputDevice &device, int format);
    ~ShmemOutboard();
    virtual int command(int argc, const char*const* argv);
    virtual void setcolor(int c) {}

 protected:
  void init_pool(void);
  void resize(int width, int height);
  ShmemOutputDevice &device_;
  CAVImagePool *imagePool_;
  int scale_;
  int lwidth_;
};

#ifdef FT_RGB
class ShmemOutboard_rgb : public ShmemOutboard
{
public:
    ShmemOutboard_rgb(ShmemOutputDevice &device);

protected:
    virtual int consume(const VideoFrame*);
};
#endif

class ShmemOutboard_yuv : public ShmemOutboard
{
public:
    ShmemOutboard_yuv(ShmemOutputDevice &device, int format)
	: ShmemOutboard(device, format) {};

protected:
    virtual int consume(const VideoFrame*);
    virtual void yuv_to_rgb(unsigned char *frame,
			    unsigned long ts) = 0;
};

class ShmemOutboard_422 : public ShmemOutboard_yuv
{
public:
    ShmemOutboard_422(ShmemOutputDevice &device);

protected:
    virtual void yuv_to_rgb(unsigned char *frame,
			    unsigned long ts);
};

class ShmemOutboard_411 : public ShmemOutboard_yuv
{
public:
    ShmemOutboard_411(ShmemOutputDevice &device);

protected:
    virtual void yuv_to_rgb(unsigned char *frame, unsigned long ts);
};

static ShmemOutputDevice shmem_device("cave");

ShmemOutputDevice::ShmemOutputDevice(const char *name) :
			OutputDevice(name)
{

}

void ShmemOutputDevice::attach_shmem(int id)
{
  cshm_ = new FLSharedMemory(id);
}

int ShmemOutputDevice::command(int argc, const char*const* argv)
{
    Tcl& tcl = Tcl::instance();
#ifdef OSDEBUG
    printf("ShmemOutputDevice::command argc=%d ", argc);
    for (int i = 0; i < argc; i++)
	printf("\"%s\" ", argv[i]);
    printf("\n");
#endif /* OSDEBUG */

    if (argc == 3) {
	if (strcmp(argv[1], "assistor") == 0)
	{
#ifdef FT_RGB
	    if (strcmp(argv[2], "rgb") == 0)
	    {
		printf("Assist with rgb\n");
		ShmemOutboard *p = new ShmemOutboard_rgb(*this);
		tcl.result(p->name());
		return (TCL_OK);
	    }
#endif
	}

	if (strcmp(argv[1], "renderer") == 0) {
	    if (strcmp(argv[2], "422") == 0) {

		ShmemOutboard* p = new ShmemOutboard_422(*this);
		tcl.result(p->name());
		return (TCL_OK);
	    }

	    if (strcmp(argv[2], "411") == 0) {
		ShmemOutboard* p = new ShmemOutboard_411(*this);
		tcl.result(p->name());
		return (TCL_OK);
	    }
	}

    }

    return (OutputDevice::command(argc, argv));
}

ShmemOutboard::ShmemOutboard(ShmemOutputDevice &device, int format) :
    				Renderer(format), device_(device),
				imagePool_(0)
{
    printf("Construct ShmemOutboard for format %d\n", format);

    imagePool_ = device_.get_pool();
}

ShmemOutboard::~ShmemOutboard()
{
    printf("Destruct ShmemOutboard()\n");
}

void ShmemOutboard::init_pool(void)
{
  imagePool_ = device_.get_pool();
  if(imagePool_ != NULL) 
    {
      imagePool_->_nextImage = -1;
      imagePool_->_currentSerial = 0;
      imagePool_->_clientActive = 1;
    }
}

int ShmemOutboard::command(int argc, const char*const* argv)
{
    Tcl& tcl = Tcl::instance();
#ifdef OSDEBUG
    printf("ShmemOutboard::command argc=%d ", argc);
    for (int i = 0; i < argc; i++)
	printf("\"%s\" ", argv[i]);
    printf("\n");
#endif /* OSDEBUG */

    if (argc == 3) {
	if (strcmp(argv[1], "scale") == 0) {
	    int s = atoi(argv[2]);
	    if (scale_ != s) {
		scale_ = s;
		resize(width_, height_);
	    }
	    return (TCL_OK);
	}
	if (strcmp(argv[1], "id") == 0)
	  {
	    device_.attach_shmem(atoi(argv[2]));
	    init_pool();
	    return (TCL_OK);
	  }
	/*
	if (strcmp(argv[1], "port") == 0) {
	    port_ = device_.lookup_port(argv[2]);
	    startup_vl();
	    return (TCL_OK);
	}
	*/
    }
    return (Renderer::command(argc, argv));
}

int ShmemOutboard_yuv::consume(const VideoFrame* frame)
{
    if (!samesize(frame))
      resize(frame->width_, frame->height_);
    
    yuv_to_rgb(frame->bp_, frame->ts_);
	       
    return 0;
}

void ShmemOutboard::resize(int w, int h)
{
    width_ = w;
    height_ = h;
    framesize_ = w * h;

    if (width_ < TEX_WIDTH)
    {
	lwidth_ = width_;
    }
    else
    {
	lwidth_ = TEX_WIDTH;
    }
}

#ifdef FT_RGB
ShmemOutboard_rgb::ShmemOutboard_rgb(ShmemOutputDevice &device) :
    				ShmemOutboard(device, FT_RGB)
{
    printf("Construct ShmemOutboard_rgb\n");
}

int ShmemOutboard_rgb::consume(const VideoFrame* frame)
{
    static int framect = 0;
    RGBFrame *rgb = (RGBFrame *) frame;
    printf("Consume rgb %dx%d ts=%d len=%d\n", frame->width_, frame->height_,
	frame->ts_, rgb->len_);

    memcpy(image_->data, frame->bp_, rgb->len_);
    image_->SetTimestamp(framect++);
	  
    return 1;
}
#endif

ShmemOutboard_422::ShmemOutboard_422(ShmemOutputDevice &device) :
    				ShmemOutboard_yuv(device, FT_YUV_422)
{
    printf("Construct ShmemOutboard_422\n");
}

void ShmemOutboard_422::yuv_to_rgb(unsigned char *frame, unsigned long ts)
{
    unsigned char *yp, *up, *vp;
    unsigned char *yl, *ul, *vl;
    int w, h, t;
    int framesize;
    int y, u, v;
    unsigned char r, g, b;
    static int once = 0;

    CAVImageDescriptor *image;
    volatile unsigned char *ImageData;
    int nextIndex;

    nextIndex = (imagePool_->_nextImage + 1) % imagePool_->_nImages;
    image = &(imagePool_->_images[nextIndex]);
    image->SetSize(TEX_WIDTH, TEX_HEIGHT, 3);
    ImageData = image->GetData();

    yp = frame;
    up = frame + framesize_;
    vp = up + framesize_ / 2;

    yl = yp;
    ul = up;
    vl = vp;

    for (h = 0; h < height_; h++)
    {
	yl = yp + h * width_;
	ul = up + h * width_ / 2;
	vl = vp + h * width_ / 2;
	for (w = 0; w < lwidth_; w += 2)
	{
	    u = (int) (*ul++ - 128);
	    v = (int) (*vl++ - 128);
	  
	    y = (int) (*yl++);

	    b = (unsigned char) UCLIMIT(y + u * 454 / 256);
	    g = (unsigned char) UCLIMIT(y - (u * 88 + v * 183) / 256);
	    r = (unsigned char) UCLIMIT(y + 385 * v / 256);

	    ImageData[TEX_OFFSET_R(w, h)] = r;
	    ImageData[TEX_OFFSET_G(w, h)] = g;
	    ImageData[TEX_OFFSET_B(w, h)] = b;

	    y = (int) (*yl++);

	    b = (unsigned char) UCLIMIT(y + u * 454 / 256);
	    g = (unsigned char) UCLIMIT(y - (u * 88 + v * 183) / 256);
	    r = (unsigned char) UCLIMIT(y + 385 * v / 256);

	    ImageData[TEX_OFFSET_R(w + 1, h)] = r;
	    ImageData[TEX_OFFSET_G(w + 1, h)] = g;
	    ImageData[TEX_OFFSET_B(w + 1, h)] = b;
	}
    }
    image->SetTimestamp(ts);
    imagePool_->_nextImage = nextIndex;
    imagePool_->_currentSerial++;
#ifdef OSDEBUG
    fprintf(stderr, "New frame ts=%d %dx%d\n", image_->timestamp,
	   image_->width, image_->height);
#endif
    if (once == 0)
    {
	once = 1;
	FILE *fp = fopen("img.ppm", "w");
	if (fp == NULL)
	{
	    printf("can't open img.ppm\n");
	}
	else
	{
	    fprintf(fp, "P6\n");
	    fprintf(fp, "%d %d\n", TEX_WIDTH, TEX_HEIGHT);
	    fprintf(fp, "255\n");
	    fwrite((const void *)ImageData, 3, TEX_WIDTH * TEX_HEIGHT, fp);
	    fclose(fp);
	}
    }
}

ShmemOutboard_411::ShmemOutboard_411(ShmemOutputDevice &device) :
    				ShmemOutboard_yuv(device, FT_YUV_411)
{
    printf("Construct ShmemOutboard_411\n");
}

void ShmemOutboard_411::yuv_to_rgb(unsigned char *frame, unsigned long ts)
{
    unsigned char *yp, *up, *vp;
    int x, w, h, k, t;
    int framesize;
    double y, u, v;
    unsigned char r, g, b;
    static int once = 0;
    unsigned char *yl, *ul, *vl;
    
    CAVImageDescriptor *image;
    volatile unsigned char *ImageData;
    int nextIndex;

    nextIndex = (imagePool_->_nextImage + 1) % imagePool_->_nImages;
    image = &(imagePool_->_images[nextIndex]);
    ImageData = image->GetData();

    yp = frame;
    up = frame + framesize_;
    vp = up + framesize_ / 4;

    image->SetSize(TEX_WIDTH, TEX_HEIGHT, 3);

    //new
    for (h = 0; h < height_; h += 2)
    {
	ul = up + (h / 2) * width_ / 2;
	vl = vp + (h / 2) * width_ / 2;

	// ul and vl are the uv for this and the next line

	// First line
	yl = yp + h * width_;
	
	for (w = 0; w < lwidth_; w += 2)
	{
	    u = (int) (*ul++ - 128);
	    v = (int) (*vl++ - 128);
	  
	    y = (int) (*yl++);

	    b = (unsigned char) UCLIMIT(y + u * 454 / 256);
	    g = (unsigned char) UCLIMIT(y - (u * 88 + v * 183) / 256);
	    r = (unsigned char) UCLIMIT(y + 385 * v / 256);

	    ImageData[TEX_OFFSET_R(w, h)] = r;
	    ImageData[TEX_OFFSET_G(w, h)] = g;
	    ImageData[TEX_OFFSET_B(w, h)] = b;

	    y = (int) (*yl++);

	    b = (unsigned char) UCLIMIT(y + u * 454 / 256);
	    g = (unsigned char) UCLIMIT(y - (u * 88 + v * 183) / 256);
	    r = (unsigned char) UCLIMIT(y + 385 * v / 256);

	    ImageData[TEX_OFFSET_R(w + 1, h)] = r;
	    ImageData[TEX_OFFSET_G(w + 1, h)] = g;
	    ImageData[TEX_OFFSET_B(w + 1, h)] = b;
	}

	// Next line; reset ul/vl and bump yl

	ul = up + (h / 2) * width_ / 2;
	vl = vp + (h / 2) * width_ / 2;
	yl = yp + (h + 1) * width_;
	
	for (w = 0; w < lwidth_; w += 2)
	{
	    u = (int) (*ul++ - 128);
	    v = (int) (*vl++ - 128);
	  
	    y = (int) (*yl++);

	    b = (unsigned char) UCLIMIT(y + u * 454 / 256);
	    g = (unsigned char) UCLIMIT(y - (u * 88 + v * 183) / 256);
	    r = (unsigned char) UCLIMIT(y + 385 * v / 256);

	    ImageData[TEX_OFFSET_R(w, (h+1))] = r;
	    ImageData[TEX_OFFSET_G(w, (h+1))] = g;
	    ImageData[TEX_OFFSET_B(w, (h+1))] = b;

	    y = (int) (*yl++);

	    b = (unsigned char) UCLIMIT(y + u * 454 / 256);
	    g = (unsigned char) UCLIMIT(y - (u * 88 + v * 183) / 256);
	    r = (unsigned char) UCLIMIT(y + 385 * v / 256);

	    ImageData[TEX_OFFSET_R(w + 1, (h+1))] = r;
	    ImageData[TEX_OFFSET_G(w + 1, (h+1))] = g;
	    ImageData[TEX_OFFSET_B(w + 1, (h+1))] = b;
	}
    }
    
#if 0
    for (h = 0; h < height_; h++)
    {
	x = 0;
	for (w = 0; w < width_; w += 2)
	{
	    for (k = 0; k < 2; k++)
	    {
		r = g = b = (yp[yi++]);

		if (w % 5 != 0)
		{
		    if (x < TEX_WIDTH)
		    {
			image_->data[TEX_OFFSET_R(x, h)] = r;
			image_->data[TEX_OFFSET_G(x, h)] = g;
			image_->data[TEX_OFFSET_B(x, h)] = b;
		    }
		    x++;
		}
	    }
	}
    }
#endif
    image->SetTimestamp(ts);
    imagePool_->_nextImage = nextIndex;
    imagePool_->_currentSerial++;
#ifdef OSDEBUG
    fprintf(stderr,"New 411 frame ts=%d %dx%d\n", image_->timestamp,
	   image_->width, image_->height);
#endif
    if (once == 0)
    {
	once = 1;
	static int idx = 0;
	char fn[1000];
	sprintf(fn, "img411.%d.ppm", idx++);
	printf("writing %s\n", fn);
	FILE *fp = fopen(fn, "w");
	if (fp == NULL)
	{
	    printf("can't open img.ppm\n");
	}
	else
	{
	    fprintf(fp, "P6\n");
	    fprintf(fp, "%d %d\n", TEX_WIDTH, TEX_HEIGHT);
	    fprintf(fp, "255\n");
	    fwrite((const void *)ImageData, 3, TEX_WIDTH * TEX_HEIGHT, fp);
	    fclose(fp);
	}
    }
}

